/*
	File : mkmap.c
*/

/* --- vO̐ ---

@lt@g̃p[^͂āClt@g̉摜쐬vOD
  <<Phantom Data>> iȉ~̃p[^j
        xo -> X Coordinate    iȉ~̒SxWj
        yo -> Y Coordinate    iȉ~̒SyWj
         a -> Minor Axis      iȉ~̒Z̒Gxj
         b -> Major Axis      iȉ~̒̒Gyj
        th -> Rotating angle  iȉ~̉]pxj
        de -> Density         iȉ~̔Zxj

́F
  1Dlt@g̃p[^t@CieLXgt@Cj
  2Dlt@g摜̏o̓t@C@iPx_^j
  3D摜x̃}gNXTCYi2̃xLj
  4D摜ỹ}gNXTCYi2̃xLj
  5D摜쐬̍ۂ̃TusNZ̃TCYij

óF
  lt@g̉摜

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  6     /* number of parameters + 1 */

typedef struct phan_data {   /* Phantom data */
	double  xo;     /*  X Coordinate */
	double  yo;     /*  Y Coordinate */
	double  a;      /*  Minor Axis */
	double  b;      /*  Major Axis */
	double  th;     /*  Rotation angle */
	double  de;     /*  Density */
	struct phan_data  *next; /*  next self pointer */
} PH_DATA;

typedef struct {
	char   f1[50]; /* input file name */
	char   f2[50]; /* output file name */
	int    nx;     /* number of matrix (x) */
	int    ny;     /* number of matrix (y) */
	int    sn;     /* pixel separate number */
	PH_DATA  *pd;  /* pointer of Phantom data */
} Param;

char *menu[PN] = {
	"Make T1,T2,Ro Map Images",
	"Input  file name <.pmt>  ",
	"Output file name <float> ",
	"Number of matrix  (x)    ",
	"Number of matrix  (y)    ",
	"Number of separate pixel ",
	};

void read_phantom_data(Param *);
void write_data(char *, float *, int);
void make_map(float *, Param *);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Print Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "n0.pmt");
	sprintf( pm->f2, "n1.img");
	pm->nx = 128;
	pm->ny = 128;
	pm->sn = 8;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f2 );
		if(*gets(dat) != '\0')  strcpy(pm->f2, dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->sn );
		if(*gets(dat) != '\0')  pm->sn = atoi(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) strcpy( pm->f2, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->sn = atoi( argv[i++] );
	}
	else {
		usage(argc, argv);
	}

}

void main(int argc, char *argv[] )
{
	Param   *pm;
	float   *img;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	img = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));

	printf(" *** Read Phantom data   ***\n");
	pm->pd = (PH_DATA *)malloc(sizeof(PH_DATA));
	pm->pd->next = NULL;
	read_phantom_data(pm);

	printf(" *** Making Map ***\n");
	make_map(img, pm);

	printf(" *** Write Image data   ***\n");
	write_data(pm->f2, img, pm->nx*pm->ny);

	free(img);
	free(pm);
}

void read_phantom_data(Param *pm)
{
	int     i, k, flag;
	char    dat[256];
	double  w[6];
	FILE    *fp;
	PH_DATA *now;

	/* open Phantom parameter file */
	if((fp = fopen(pm->f1, "r")) == NULL) {
		fprintf( stderr, "Error: file open [%s].\n", pm->f1);
		exit(1);
	}

	/* Input Phatom parameters */
	now = pm->pd;
	flag = 0;
	while(fgets(dat,100,fp) != NULL) {
		if(*dat=='#'){
			printf("      ");
			printf(dat);
			continue;
		}
		for(i=0;i<6;i++) w[i]=0;
		k = 0;
		for(i=0;i<6;i++){
		  while((dat[k]==' ')||(dat[k]=='\t')) k++;
		  w[i]=atof(dat+k);
		  while((dat[k]!=' ')&&(dat[k]!='\t')) k++;
		}
		if(flag) {
			now->next = (PH_DATA *)malloc(sizeof(PH_DATA));
			now = now->next;
			now->next = NULL;
		}
		now->xo = w[0];
		now->yo = w[1];
		now->a  = w[2];
		now->b  = w[3];
		now->th = w[4];
		now->de = w[5];
		flag++;
		printf("* %2d *", flag);
		printf("%8.4f,",  now->xo);
		printf("%8.4f,",  now->yo);
		printf("%8.4f,",  now->a);
		printf("%8.4f,",  now->b);
		printf("%8.4f,",  now->th);
		printf("%8.4f\n", now->de);
	}
	printf("\n");
	fclose(fp);
}

void make_map(float *img, Param *pm)
{
	int	     i, j, k, kx, ky;
	double   x, y, x1, y1, x2, y2, sub, rad, si, co;
	PH_DATA  *now;

	for(i = 0 ; i < pm->nx*pm->ny ; i++)
		img[i] = 0;
	k = 0;
	now = pm->pd;
	while(now != NULL) {
	  printf(" [%2d] ", ++k);
	  printf(" xo =%7.3f",   now->xo);
	  printf(" yo =%7.3f",   now->yo);
	  printf(" a =%7.3f",    now->a);
	  printf(" b =%7.3f",    now->b);
	  printf(" th =%7.3f",   now->th);
	  printf(" de =%7.3f\n", now->de);
	  rad = PI*now->th/180.;
	  si = sin(rad);
	  co = cos(rad);
	  for(i = 0 ; i < pm->ny ; i++) {
		y1 = (pm->ny/2-i)*2/(double)pm->ny;
	    for(j = 0 ; j < pm->nx ; j++) {
		  x1 = (j-pm->nx/2)*2/(double)pm->nx;
	      sub = 0;
	      for(ky = pm->sn/2 ; ky > -pm->sn/2 ; ky--) {
	        y2 = y1+ky/(double)(pm->sn*pm->ny)-now->yo;
	        for(kx = -pm->sn/2 ; kx < pm->sn/2 ; kx++) {
	          x2 = x1+kx/(double)(pm->nx*pm->sn)-now->xo;
	          x =  x2*co+y2*si;
	          y = -x2*si+y2*co;
	          if(x*x/(now->a*now->a)+y*y/(now->b*now->b) <= 1.)
	            sub += now->de;
	        }
          }
	      img[i*pm->nx+j] += (float)(sub/(pm->sn*pm->sn));
	    }
	  }
	  now = now->next;
	}
}

