/*
	File : point_3D_Fourier.c
*/

/* --- vO̐ ---

10́@MRI3č\V~[V
1߁@1_摜3v
2߁@1_摜3č\

@_̈ʒu3œ͂āC3t[Gϊ@̌vf[^쐬C
̃f[^3t[Gϊgčč\vOD

́F
  1Dč\ꂽ摜̏o̓t@CiPx_^j
  2D摜x̃}gNXTCYi2̃xLj
  3D摜ỹ}gNXTCYi2̃xLj
  4D摜z̃}gNXTCYi2̃xLj
  5DziGj
  6DTvOԊuimsj
  7D_ẍʒuicmj
  8D_ÿʒuicmj
  9D_z̈ʒuicmj

óF
  vf[^̎̉摜ip0.imgj
  vf[^̋̉摜ip1.imgj
  č\ꂽ_摜

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  10     /* number of parameters + 1 */


typedef struct {
	char    f1[50]; /* output file  name */
	int     nx;     /* number of matrix (x) */
	int     ny;     /* number of matrix (y) */
	int     nz;     /* number of matrix (z) */
	double  gx;     /* Gradient Magnetic Field (gauss) */
	double  sm;     /* Sampling time (ms) */
	double  px;     /* x-coordinate of point source */
	double  py;     /* y-coordinate of point source */
	double  pz;     /* z-coordinate of point source */
	float   *imr;   /* Image data (real)  */
	float   *imi;   /* Image data (imaginary) */
} Param;

char *menu[PN] = {
	"MRI reconstruction (3D-Fourier Method) for point source",
	"Output file name <.img>           ",
	"Number of matrix  (x)             ",
	"Number of matrix  (y)             ",
	"Number of matrix  (z)             ",
	"Gradient Magnetic field (G)       ",
	"Sampling time (ms)                ",
	"x-coordinate of point source (cm) ",
	"y-coordinate of point source (cm) ",
	"z-coordinate of point source (cm) ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void write_data3d(char *, float *, int, int);
void make_point(Param *);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void FFT_3D(int, float *, float *, int, int, int);


void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Print Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "n");
	pm->nx = 128;
	pm->ny = 128;
	pm->nz = 128;
	pm->gx = 0.12;
	pm->sm = 0.1;
	pm->px = 2.0;
	pm->py = 1.0;
	pm->pz = 3.0;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nz );
		if(*gets(dat) != '\0')  pm->nz = atoi(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->gx );
		if(*gets(dat) != '\0')  pm->gx = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->sm );
		if(*gets(dat) != '\0')  pm->sm = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->px );
		if(*gets(dat) != '\0')  pm->px = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->py );
		if(*gets(dat) != '\0')  pm->py = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->pz );
		if(*gets(dat) != '\0')  pm->pz = atof(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->nz = atoi( argv[i++] );
		if((argc--) > 1) pm->gx = atof( argv[i++] );
		if((argc--) > 1) pm->sm = atof( argv[i++] );
		if((argc--) > 1) pm->px = atof( argv[i++] );
		if((argc--) > 1) pm->py = atof( argv[i++] );
		if((argc--) > 1) pm->pz = atof( argv[i++] );
	}
	else {
		usage(argc, argv);
	}

}

void main(int argc, char *argv[] )
{
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->imr = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nz*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nz*sizeof(float));

	printf(" *** Making Point Image ***\n");
	make_point(pm);
	write_data3d("p0", pm->imr, pm->nx*pm->ny, pm->nz);
	write_data3d("p1", pm->imi, pm->nx*pm->ny, pm->nz);

	printf(" *** 3D-IFT  ***\n");
	FFT_3D(-1, pm->imr, pm->imi, pm->nx, pm->ny, pm->nz);

	printf(" *** Write Image data   ***\n");
	write_data3d(pm->f1, pm->imr, pm->nx*pm->ny, pm->nz);

	free(pm->imr);
	free(pm->imi);
	free(pm);
}

void make_point(Param *pm)
{
	int	     k, l, m;
	double   th, g0, g1, h=1000;

	/*
		(10-10a)(10-10b)(10-10c)
		x̌zωƂy̌z
		鎞Ԃ킹邽߂ɁCg1ɂ܂Ƃ߂ĂD
	*/
	g0 = 2*PI*42.58*1e6;
	g1 = pm->gx*1e-4*pm->sm*1e-3;
	for(m = 0 ; m < pm->nz ; m++) {
		for(l = 0 ; l < pm->ny ; l++) {
			for(k = 0 ; k < pm->nx ; k++) {
				th = g0*g1*(pm->px*(k-pm->nx/2)+pm->py*(pm->ny/2-l)+pm->pz*(pm->nz/2-m));
				pm->imr[m*pm->ny*pm->nx+l*pm->nx+k] = (float)(h*cos(th));
				pm->imi[m*pm->ny*pm->nx+l*pm->nx+k] = (float)(h*sin(th));
			}
		}
	}
}
