/*
	File : point_FBP.c
*/

/* --- vO̐ ---

5́@eč\̃V~[V
2߁@1_摜̍č\

@_̈ʒu͂āCeč\@̌vf[^쐬C
̃f[^FBPgčč\vOD

́F
  1Dč\ꂽ摜̏o̓t@CiPx_^j
  2D摜x̃}gNXTCYi2̃xLj
  3D摜ỹ}gNXTCYi2̃xLj
  4D_̈ʒuicmj
  5DziGj
  6DTvOԊuimsj

óF
  vf[^̎̉摜ip0.imgj
  vf[^̋̉摜ip1.imgj
  č\ꂽ_摜

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj
  filter.c    iFBPp̃tB^֐̃t@Cj
  backproj.c  ite֐̃t@Cj

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  7     /* number of parameters + 1 */


typedef struct {
	char    f1[50]; /* output file  name */
	int     nx;     /* number of matrix (x) */
	int     ny;     /* number of matrix (y) */
	double  ps;     /* Point source (cm) */
	double  gm;     /* Gradient Magnetic Field (gauss) */
	double  sm;     /* Sampling time (ms) */
	float   *imr;   /* Image data (real)  */
	float   *imi;   /* Image data (imaginary) */
	float   *img;   /* Reconstructed Image data */
} Param;

char *menu[PN] = {
	"MRI reconstruction (FBP) for point source",
	"Output file name <.img>     ",
	"Number of matrix  (x)       ",
	"Number of matrix  (y)       ",
	"Point Source (cm)           ",
	"Gradient Magnetic field (G) ",
	"Sampling time (ms)          ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void make_point(Param *);
void proj_ift(Param *);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void filter100(float *, float *, int, double);
void BackProjection(int,float *,int,int,float *,int,int);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Print Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "n0.img");
	pm->nx = 128;
	pm->ny = 128;
	pm->ps = 5.0;
	pm->gm = 0.1;
	pm->sm = 0.1;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->ps );
		if(*gets(dat) != '\0')  pm->ps = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->gm );
		if(*gets(dat) != '\0')  pm->gm = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->sm );
		if(*gets(dat) != '\0')  pm->sm = atof(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->ps = atof( argv[i++] );
		if((argc--) > 1) pm->gm = atof( argv[i++] );
		if((argc--) > 1) pm->sm = atof( argv[i++] );
	}
	else {
		usage(argc, argv);
	}

}

void main(int argc, char *argv[] )
{
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->imr = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->img = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));

	printf(" *** Making Point Source Date ***\n");
	make_point(pm);
	write_data("p0.img", pm->imr, pm->nx*pm->ny);
	write_data("p1.img", pm->imi, pm->nx*pm->ny);

	printf(" *** Filtering and inverse FFT  ***\n");
	proj_ift(pm);

	printf(" *** Back-Projection  ***\n");
	BackProjection(1, pm->img, pm->nx, pm->ny, pm->imr, pm->nx, pm->ny);

	printf(" *** Write Image data   ***\n");
	write_data(pm->f1, pm->img, pm->nx*pm->ny);

	free(pm->imr);
	free(pm->imi);
	free(pm->img);
	free(pm);
}

void make_point(Param *pm)
{
	int	     i, j;
	double   co, th;

	for(i = 0 ; i < pm->ny ; i++) {
		co = cos(i*PI/(double)pm->ny);
		for(j = 0 ; j < pm->nx ; j++) {
			/*
				(5-12a)(5-12b)
			*/
			th = 2*PI*42.58*1e6*pm->gm*1e-4*pm->ps*co*pm->sm*1e-3*j;
			pm->imr[i*pm->nx+j] = (float)cos(th);
			pm->imi[i*pm->nx+j] = (float)-sin(th);
		}
	}
}

void proj_ift(Param *pm)
{
	int    i, j;
	float  *xr, *xi, *si, *co, buff;
	unsigned short *brv;

	si = (float *)malloc(pm->nx*sizeof(float)/2);
	co = (float *)malloc(pm->nx*sizeof(float)/2);
	brv = (unsigned short *)malloc(pm->nx*sizeof(unsigned short));

	FFTInit(pm->nx, si, co, brv);

	for(i = 0 ; i < pm->ny ; i++) {
		xr = pm->imr+i*pm->nx;
		xi = pm->imi+i*pm->nx;
		filter100(xr, xi, pm->nx, 20./128.);
		FFT(1, pm->nx, xr, xi, si, co, brv);
		for(j = 0 ; j < pm->nx/2 ; j++) {
			buff = xr[j];
			xr[j] = xr[j+pm->nx/2];
			xr[j+pm->nx/2] = buff;
		}
	}
	free(si);
	free(co);
	free(brv);
}
