/*
	File : point_motion_Fourier.c
*/

/* --- vO̐ ---

8́@MRIɂ郂[VA[`t@Ng
2߁@2t[Gϊ@ɂ郂[VA[`t@Ng̉e
 (1)   1_摜̃V~[V

@_̈ʒuƁuv͂āC2t[Gϊ@̌vf[^쐬C
̃f[^2t[Gϊgčč\vOD

́F
  1Dč\ꂽ摜̏o̓t@CiPx_^j
  2D摜x̃}gNXTCYi2̃xLj
  3D摜ỹ}gNXTCYi2̃xLj
  4DziGj
  5DTvOԊuimsj
  6D_ẍʒuicmj
  7D_ÿʒuicmj
  8. ̕i0:xC1:yj
  9. öʒu
 10. Iʒu
 11. ̑isNZ/msj

óF
  vf[^̎̉摜ip0.imgj
  vf[^̋̉摜ip1.imgj
  č\ꂽ_摜

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  12     /* number of parameters + 1 */


typedef struct {
	char    f1[50]; /* output file  name */
	int     nx;     /* number of matrix (x) */
	int     ny;     /* number of matrix (y) */
	double  gx;     /* Gradient Magnetic Field (gauss) */
	double  sm;     /* Sampling time (ms) */
	double  px;     /* x-coordinate of point source */
	double  py;     /* y-coordinate of point source */
	int     mxy;    /* motion x or y */
	int     sp;     /* start point of move */
	int     ep;     /* end   point of move */
	double  v;      /* velocity of move (pixel/ms) */
	float   *imr;   /* Image data (real)  */
	float   *imi;   /* Image data (imaginary) */
} Param;

char *menu[PN] = {
	"MRI reconstruction (Fourier Method) for point source with motion",
	"Output file name <.img>           ",
	"Number of matrix  (x)             ",
	"Number of matrix  (y)             ",
	"Gradient Magnetic field (G)       ",
	"Sampling time (ms)                ",
	"x-coordinate of point source (cm) ",
	"y-coordinate of point source (cm) ",
	"motion x or y ? (x:0, y:1)        ",
	"start point of move               ",
	"end   point of move               ",
	"velocity of move (pixel/ms)       ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void make_point_motionx(Param *);
void make_point_motiony(Param *);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void FFT_2D(int, float *, float *, int, int);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Print Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "n0.img");
	pm->nx = 128;
	pm->ny = 128;
	pm->gx = 0.12;
	pm->sm = 0.1;
	pm->px = 0.5;
	pm->py = 0.3;
	pm->mxy= 0;
	pm->sp = -10;
	pm->ep = 10;
	pm->v  = 0.15;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->gx );
		if(*gets(dat) != '\0')  pm->gx = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->sm );
		if(*gets(dat) != '\0')  pm->sm = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->px );
		if(*gets(dat) != '\0')  pm->px = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->py );
		if(*gets(dat) != '\0')  pm->py = atof(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->mxy);
		if(*gets(dat) != '\0')  pm->mxy= atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->sp );
		if(*gets(dat) != '\0')  pm->sp = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->ep );
		if(*gets(dat) != '\0')  pm->ep = atoi(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->v );
		if(*gets(dat) != '\0')  pm->v = atof(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->gx = atof( argv[i++] );
		if((argc--) > 1) pm->sm = atof( argv[i++] );
		if((argc--) > 1) pm->px = atof( argv[i++] );
		if((argc--) > 1) pm->py = atof( argv[i++] );
		if((argc--) > 1) pm->mxy= atoi( argv[i++] );
		if((argc--) > 1) pm->sp = atoi( argv[i++] );
		if((argc--) > 1) pm->ep = atoi( argv[i++] );
		if((argc--) > 1) pm->v  = atof( argv[i++] );
	}
	else {
		usage(argc, argv);
	}

	/* Error */
	if(pm->sp > pm->ep) {
		fprintf(stderr, "Error : [move] start point > end point \n");
		exit(1);
	}
}

void main(int argc, char *argv[] )
{
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->imr = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));

	printf(" *** Making Point Image ***\n");
	if(!pm->mxy) {
		printf(" --- motion x ---\n");
		make_point_motionx(pm);
	}
	else {
		printf(" --- motion y ---\n");
		make_point_motiony(pm);
	}
	write_data("p0.img", pm->imr, pm->nx*pm->ny);
	write_data("p1.img", pm->imi, pm->nx*pm->ny);

	printf(" *** 2D-IFT  ***\n");
	FFT_2D(-1, pm->imr, pm->imi, pm->nx, pm->ny);

	printf(" *** Write Image data   ***\n");
	write_data(pm->f1, pm->imr, pm->nx*pm->ny);

	free(pm->imr);
	free(pm->imi);
	free(pm);
}

void make_point_motionx(Param *pm)
// x̓
{
	int	     i, j;
	double   th, g0, g1; // [sp:0, ep:0, v:50.];[sp:-9, ep:10, v:0.1];

	/*
		(8-44)(8-49)(8-52)(8-55)
		x̌zωƂy̌z
		鎞Ԃ킹邽߂ɁCg1ɂ܂Ƃ߂ĂD
	*/
	g0 = 2*PI*42.58*1e6;
	g1 = pm->gx*1e-4*pm->sm*1e-3;
	for(i = 0 ; i < pm->ny/2+pm->sp ; i++) { // before movement
		for(j = 0 ; j < pm->nx ; j++) {
			th = g0*g1*(pm->px*(j-pm->nx/2)+pm->py*(pm->ny/2-i));
			pm->imr[i*pm->nx+j] = (float)cos(th);
			pm->imi[i*pm->nx+j] = (float)sin(th);
		}
	}
	for(i = pm->ny/2+pm->sp ; i <= pm->ny/2+pm->ep ; i++) { // midst of movement
		for(j = 0 ; j < pm->nx ; j++) {
			th = g0*g1*((pm->px+pm->v*(j-pm->nx/2)*1e-3*pm->sm+(i-pm->ny/2-pm->sp)*pm->v)*(j-pm->nx/2)+pm->py*(pm->ny/2-i));
			pm->imr[i*pm->nx+j] = (float)cos(th);
			pm->imi[i*pm->nx+j] = (float)sin(th);
		}
	}
	for(i = pm->ny/2+pm->ep+1 ; i < pm->ny ; i++) { // after movement
		for(j = 0 ; j < pm->nx ; j++) {
			th = g0*g1*((pm->px+pm->v*(pm->ep-pm->sp+1))*(j-pm->nx/2)+pm->py*(pm->ny/2-i));
			pm->imr[i*pm->nx+j] = (float)cos(th);
			pm->imi[i*pm->nx+j] = (float)sin(th);
		}
	}
}

void make_point_motiony(Param *pm)
// y̓
{
	int	     i, j;
	double   th, g0, g1;

	/*
		(8-50)(8-51)(8-56)(8-59)
		x̌zωƂy̌z
		鎞Ԃ킹邽߂ɁCg1ɂ܂Ƃ߂ĂD
	*/
	g0 = 2*PI*42.58*1e6;
	g1 = pm->gx*1e-4*pm->sm*1e-3;
	for(i = 0 ; i < pm->ny/2+pm->sp ; i++) { // before movement
		for(j = 0 ; j < pm->nx ; j++) {
			th = g0*g1*(pm->px*(j-pm->nx/2)+pm->py*(pm->ny/2-i));
			pm->imr[i*pm->nx+j] = (float)cos(th);
			pm->imi[i*pm->nx+j] = (float)sin(th);
		}
	}
	for(i = pm->ny/2+pm->sp ; i <= pm->ny/2+pm->ep ; i++) { // midst of movement
		for(j = 0 ; j < pm->nx ; j++) {
			th = g0*g1*(pm->px*(j-pm->nx/2)+(pm->py+(i-pm->ny/2-pm->sp)*pm->v)*(pm->ny/2-i));
			pm->imr[i*pm->nx+j] = (float)cos(th);
			pm->imi[i*pm->nx+j] = (float)sin(th);
		}
	}
	for(i = pm->ny/2+pm->ep+1 ; i < pm->ny ; i++) { // after movement
		for(j = 0 ; j < pm->nx ; j++) {
			th = g0*g1*(pm->px*(j-pm->nx/2)+(pm->py+pm->v*(pm->ep-pm->sp+1))*(pm->ny/2-i));
			pm->imr[i*pm->nx+j] = (float)cos(th);
			pm->imi[i*pm->nx+j] = (float)sin(th);
		}
	}
}
