/*
	File : recon_3D_Fourier.c
*/

/* --- vO̐ ---

10́@MRI3č\V~[V
3߁@lt@g̃V~[V

@vOŒ3lt@gw肵āC3t[Gϊ@̌vf[^쐬C
̃f[^3t[Gϊgčč\vOD

́F
  1Dč\ꂽ摜̏o̓t@C@iPx_^j
  2D摜x̃}gNXTCYi2̃xLj
  3D摜ỹ}gNXTCYi2̃xLj
  4D摜z̃}gNXTCYi2̃xLj
  5DziGj
  6DTvOԊuimsj

óF
  vf[^̎̉摜ip0.imgj
  vf[^̋̉摜ip1.imgj
  č\ꂽ_摜

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  7     /* number of parameters + 1 */


typedef struct {
	char    f1[50]; /* input file name */
	char    f2[50]; /* output file  name */
	int     nx;     /* number of matrix (x) */
	int     ny;     /* number of matrix (y) */
	int     nz;     /* number of matrix (z) */
	double  gm;     /* Gradient Magnetic Field (gauss) */
	double  sm;     /* Sampling time (ms) */
	float   *imr;   /* Image data (real)  */
	float   *imi;   /* Image data (imaginary) */
	float   *img;   /* Source data */
} Param;

char *menu[PN] = {
	"MRI 3D reconstruction (Fourier Method)",
	"Output file name <.img>           ",
	"Number of matrix  (x)             ",
	"Number of matrix  (y)             ",
	"Number of matrix  (z)             ",
	"Gradient Magnetic field (G)       ",
	"Sampling time (ms)                ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void write_data3d(char *, float *, int, int);
void make_source(Param *);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void FFT_3D(int, float *, float *, int, int, int);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Print Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "shepp3n.img");
	sprintf( pm->f2, "n");
	pm->nx = 128;
	pm->ny = 128;
	pm->nz = 128;
	pm->gm = 0.12;
	pm->sm = 0.1;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f2 );
		if(*gets(dat) != '\0')  strcpy(pm->f2, dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nz );
		if(*gets(dat) != '\0')  pm->nz = atoi(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->gm );
		if(*gets(dat) != '\0')  pm->gm = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->sm );
		if(*gets(dat) != '\0')  pm->sm = atof(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f2, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->nz = atoi( argv[i++] );
		if((argc--) > 1) pm->gm = atof( argv[i++] );
		if((argc--) > 1) pm->sm = atof( argv[i++] );
	}
	else {
		usage(argc, argv);
	}

}

void main(int argc, char *argv[] )
{
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->imr = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nz*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nz*sizeof(float));

	printf(" *** Make MRI source data ***\n");
	make_source(pm);
	write_data3d("p0", pm->imr, pm->nx*pm->ny, pm->nz);
	write_data3d("p1", pm->imi, pm->nx*pm->ny, pm->nz);

	printf(" *** 3D-IFT  ***\n");
	FFT_3D(-1, pm->imr, pm->imi, pm->nx, pm->ny, pm->nz);

	printf(" *** Write Image data   ***\n");
	write_data3d(pm->f2, pm->imr, pm->nx*pm->ny, pm->nz);

	free(pm->imr);
	free(pm->imi);
	free(pm);
}

void make_source(Param *pm)
{
	int	     k, l, m, i, ix, iy, iz;
	double   x, y, z, th, g0, g1, h;
	double   a1= 6.*20./pm->nx, b1=10.*20./pm->ny, c1=4.*20./pm->nz;
	double   a2= 3.*20./pm->nx, b2= 3.*20./pm->ny, c2=2.*20./pm->nz;
	double   y0= 2.*20./pm->ny;

	for(i = 0 ; i < pm->nx*pm->ny*pm->nz ; i++)
		pm->imr[i] = pm->imi[i] = 0;

	/*
		a1, b1, c1a2, b2, c2, y03ȉ~̂̃f[^
		a : x̒
		b : y̒
		c : z̒
		y0: yւ̂
		(10-12a)(10-12b)(10-12c)
		x̌zωƂy̌z
		鎞Ԃ킹邽߂ɁCg1ɂ܂Ƃ߂ĂD
	*/
	g0 = 2*PI*42.58*1e6;
	g1 = pm->gm*1e-4*pm->sm*1e-3;
	for(iz = 0 ; iz < pm->nz ; iz++) {
	z = (pm->nz/2-iz)*20./pm->nz;
	for(iy = 0 ; iy < pm->ny ; iy++) {
	y = (pm->ny/2-iy)*20./pm->ny;
	for(ix = 0 ; ix < pm->nx ; ix++) {
	x = (ix-pm->nx/2)*20./pm->nx;
	if(x*x/(a1*a1)+y*y/(b1*b1)+z*z/(c1*c1) > 1.) continue;
	printf("\r Make Point [x:%3d, y:%3d, z:%3d]", ix, iy, iz);
	h = 1;
	if(x*x/(a2*a2)+(y-y0)*(y-y0)/(b2*b2)+z*z/(c2*c2) <= 1.) h = 2;
	for(m = 0 ; m < pm->nz ; m++) {
		for(l = 0 ; l < pm->ny ; l++) {
			for(k = 0 ; k < pm->nx ; k++) {
				th = g0*g1*(x*(k-pm->nx/2)+y*(pm->ny/2-l)+z*(pm->nz/2-m));
				pm->imr[m*pm->ny*pm->nx+l*pm->nx+k] += (float)(h*cos(th));
				pm->imi[m*pm->ny*pm->nx+l*pm->nx+k] += (float)(h*sin(th));
				}
			}
		}
	}}}
	printf("\n");

}
