/*
	File : recon_Fourier.c
*/

/* --- vO̐ ---

6́@2t[Gϊ@̃V~[V
4߁@lt@g̉摜č\

@lt@g͂āC2t[Gϊ@̌vf[^쐬C
̃f[^2t[Gϊgčč\vOD

́F
  1Dlt@g摜̓̓t@CiPx_^j
  2Dč\ꂽ摜̏o̓t@C@iPx_^j
  3D摜x̃}gNXTCYi2̃xLj
  4D摜ỹ}gNXTCYi2̃xLj
  5DziGj
  6DTvOԊuimsj

óF
  vf[^̎̉摜ip0.imgj
  vf[^̋̉摜ip1.imgj
  fgĔ쐬̉摜ip2.imgj
  fgĔ쐬̉摜ip3.imgj
  č\ꂽlt@g摜

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  7     /* number of parameters + 1 */


typedef struct {
	char    f1[50]; /* input file name */
	char    f2[50]; /* output file  name */
	int     nx;     /* number of matrix (x) */
	int     ny;     /* number of matrix (y) */
	double  gm;     /* Gradient Magnetic Field (gauss) */
	double  sm;     /* Sampling time (ms) */
	float   *imr;   /* Image data (real)  */
	float   *imi;   /* Image data (imaginary) */
	float   *img;   /* Source data */
} Param;

char *menu[PN] = {
	"MRI reconstruction (Fourier Method) for phantom image",
	"Input  file name <.img>           ",
	"Output file name <.img>           ",
	"Number of matrix  (x)             ",
	"Number of matrix  (y)             ",
	"Gradient Magnetic field (G)       ",
	"Sampling time (ms)                ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void make_source(Param *);
void complex_conjucate_1D(float *, float *, int, int);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void FFT_2D(int, float *, float *, int, int);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Print Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "shepp3n.img");
	sprintf( pm->f2, "n0.img");
	pm->nx = 128;
	pm->ny = 128;
	pm->gm = 0.1;
	pm->sm = 0.1;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f2 );
		if(*gets(dat) != '\0')  strcpy(pm->f2, dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->gm );
		if(*gets(dat) != '\0')  pm->gm = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->sm );
		if(*gets(dat) != '\0')  pm->sm = atof(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) strcpy( pm->f2, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->gm = atof( argv[i++] );
		if((argc--) > 1) pm->sm = atof( argv[i++] );
	}
	else {
		usage(argc, argv);
	}

}

void main(int argc, char *argv[] )
{
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->imr = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->img = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));

	printf(" *** Read Image ***\n");
	read_data(pm->f1, pm->img, pm->nx*pm->ny);

	printf(" *** Make MRI source data ***\n");
	make_source(pm);
	write_data("p0.img", pm->imr, pm->nx*pm->ny);
	write_data("p1.img", pm->imi, pm->nx*pm->ny);

	printf(" *** Complex Conjugate ***\n");
	complex_conjucate_1D(pm->imr, pm->imi, pm->nx, pm->ny);
	write_data("p2.img", pm->imr, pm->nx*pm->ny);
	write_data("p3.img", pm->imi, pm->nx*pm->ny);

	printf(" *** 2D-IFT  ***\n");
	FFT_2D(-1, pm->imr, pm->imi, pm->nx, pm->ny);

	printf(" *** Write Image data   ***\n");
	write_data(pm->f2, pm->imr, pm->nx*pm->ny);

	free(pm->img);
	free(pm->imr);
	free(pm->imi);
	free(pm);
}

void make_source(Param *pm)
// lt@ǧvf[^쐬
{
	int	     i, j, k, m;
	double   th, g0, g1, ii, jj;

	for(i = 0 ; i < pm->nx*pm->ny ; i++)
		pm->imr[i] = pm->imi[i] = 0;

	/*
		(6-12a)(6-12b)
		x̌zωƂy̌z
		鎞Ԃ킹邽߂ɁCg1ɂ܂Ƃ߂ĂD
	*/
	g0 = 2*PI*42.58*1e6;
	g1 = pm->gm*1e-4*pm->sm*1e-3;
	for(j = 0 ; j < pm->ny ; j++) {
		jj = (pm->ny/2-j)*20./pm->ny;
		for(i = 0 ; i < pm->nx ; i++) {
			if(pm->img[j*pm->nx+i] == 0.0) continue;
			ii = (i-pm->nx/2)*20./pm->nx;
			for(m = 0 ; m < pm->ny ; m++) {
				for(k = 0 ; k < pm->nx ; k++) {
					th = g0*g1*(ii*k+jj*(pm->ny/2-m));
					pm->imr[m*pm->nx+k] += (float)(pm->img[j*pm->nx+i]*cos(th));
					pm->imi[m*pm->nx+k] += (float)(pm->img[j*pm->nx+i]*sin(th));
				}
			}
		}
	}
}

void complex_conjucate_1D(float *imr, float *imi, int nx, int ny)
{
	int    i, j;
	float  sr, si;
	void   fswap(float *, float *);

	for(i = 0 ; i < ny ; i++) {
		sr = imr[i*nx+nx/2];
		si = imi[i*nx+nx/2];
		for(j = 0 ; j < nx/2 ; j++) {
			imr[i*nx+j+nx/2] = imr[i*nx+j];
			imi[i*nx+j+nx/2] = imi[i*nx+j];
		}
		imr[i*nx] = sr;
		imi[i*nx] = -si;
	}
	for(j = 1 ; j < nx/2 ; j++) {
		imr[j] =  imr[nx-j];
		imi[j] = -imi[nx-j];
	}
	for(i = 1 ; i < nx/2 ; i++) {
		fswap(imr+i*nx, imr+(nx-i)*nx);
		fswap(imi+i*nx, imi+(nx-i)*nx);
	}
	for(i = 1 ; i < ny ; i++) {
		for(j = 1 ; j < nx/2 ; j++) {
			imr[i*nx+j] =  imr[(nx-i)*nx+nx-j];
			imi[i*nx+j] = -imi[(nx-i)*nx+nx-j];
		}
	}
}
