/*
	File : recon_SpinEcho.c
*/

/* --- vO̐ ---

7́@XsGR[@̃V~[V
2߁@lt@ǧvƍč\
3߁@p[^ςꍇ̍č\

@lt@g̐fqxT1ɘaԂT2ɘaԂ3̃t@C͂āC
2t[Gϊ@̌vf[^쐬C̃f[^2t[Gϊg
č\vOD

́F
  1Dlt@g̐fqx摜̓̓t@CiPx_^j
  2Dlt@gT1ɘaԉ摜̓̓t@CiPx_^j
     iɘaԂms̐lŉ摜́j
  3Dlt@gT2ɘaԉ摜̓̓t@CiPx_^j
     iɘaԂms̐lŉ摜́j
  4Dč\ꂽ摜̏o̓t@C@iPx_^j
  5D摜x̃}gNXTCYi2̃xLj
  6D摜ỹ}gNXTCYi2̃xLj
  7DziGj
  8DTvOԊuimsj
  9DGR[ԁ@iTEjimsj
 10DJԂԁiTRjimsj

óF
  vf[^̎̉摜ip0.imgj
  vf[^̋̉摜ip1.imgj
  č\ꂽlt@g摜

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  11     /* number of parameters + 1 */


typedef struct {
	char    f1[50]; /* input ro file name */
	char    f2[50]; /* input t1 file name */
	char    f3[50]; /* input t2 file name */
	char    f4[50]; /* output file  name */
	int     nx;     /* number of matrix (x) */
	int     ny;     /* number of matrix (y) */
	double  gm;     /* Gradient Magnetic Field (gauss) */
	double  sm;     /* Sampling time (ms) */
	double  te;     /* TE (Echo Time) */
	double  tr;     /* TR (Repeat Time ) */
	float   *imr;   /* Image data (real)  */
	float   *imi;   /* Image data (imaginary) */
	float   *iro;   /* Ro data */
	float   *it1;   /* T1 data */
	float   *it2;   /* T2 data */
} Param;

char *menu[PN] = {
	"MRI reconstruction (SpinEcho Method) for phantom image",
	"Input Ro file name <ro.img> ",
	"Input T1 file name <t1.img> ",
	"Input T2 file name <t2.img> ",
	"Output file name <.img>     ",
	"Number of matrix  (x)       ",
	"Number of matrix  (y)       ",
	"Gradient Magnetic field (G) ",
	"Sampling time (ms)          ",
	"Echo time   TE (ms)         ",
	"Repeat time TR (ms)         ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void make_source(Param *);
void complex_conjucate_1D(float *, float *, int, int);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void FFT_2D(int, float *, float *, int, int);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Print Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "shepp3ro.img");
	sprintf( pm->f2, "shepp3t1.img");
	sprintf( pm->f3, "shepp3t2.img");
	sprintf( pm->f4, "n0.img");
	pm->nx = 128;
	pm->ny = 128;
	pm->gm = 0.12;
	pm->sm = 0.1;
	pm->te = 50.;
	pm->tr = 1000.;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f2 );
		if(*gets(dat) != '\0')  strcpy(pm->f2, dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f3 );
		if(*gets(dat) != '\0')  strcpy(pm->f3, dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f4 );
		if(*gets(dat) != '\0')  strcpy(pm->f4, dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%i] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->gm );
		if(*gets(dat) != '\0')  pm->gm = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->sm );
		if(*gets(dat) != '\0')  pm->sm = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->te );
		if(*gets(dat) != '\0')  pm->te = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->tr );
		if(*gets(dat) != '\0')  pm->tr = atof(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) strcpy( pm->f2, argv[i++] );
		if((argc--) > 1) strcpy( pm->f3, argv[i++] );
		if((argc--) > 1) strcpy( pm->f4, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->gm = atof( argv[i++] );
		if((argc--) > 1) pm->sm = atof( argv[i++] );
		if((argc--) > 1) pm->te = atof( argv[i++] );
		if((argc--) > 1) pm->tr = atof( argv[i++] );
	}
	else {
		usage(argc, argv);
	}

}

void main(int argc, char *argv[] )
{
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->imr = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->iro = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->it1 = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->it2 = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));

	printf(" *** Read Image ***\n");
	read_data(pm->f1, pm->iro, pm->nx*pm->ny);
	read_data(pm->f2, pm->it1, pm->nx*pm->ny);
	read_data(pm->f3, pm->it2, pm->nx*pm->ny);

	printf(" *** Make MRI source data ***\n");
	make_source(pm);
	write_data("p0.img", pm->imr, pm->nx*pm->ny);
	write_data("p1.img", pm->imi, pm->nx*pm->ny);

	printf(" *** 2D-IFT  ***\n");
	FFT_2D(-1, pm->imr, pm->imi, pm->nx, pm->ny);

	printf(" *** Write Image data   ***\n");
	write_data(pm->f4, pm->imr, pm->nx*pm->ny);

	free(pm->imr);
	free(pm->imi);
	free(pm->iro);
	free(pm->it1);
	free(pm->it2);
	free(pm);
}

void make_source(Param *pm)
{
	int	     i, j, ix, iy;
	double   x, y, th, g0, g1;
	double   a, b, k, m;

	for(i = 0 ; i < pm->nx*pm->ny ; i++)
		pm->imr[i] = pm->imi[i] = 0;

	/*
		(7-19)(7-20)
		x̌zωƂy̌z
		鎞Ԃ킹邽߂ɁCg1ɂ܂Ƃ߂ĂD
	*/
	g0 = 2*PI*42.58*1e6;
	g1 = pm->gm*1e-4*pm->sm*1e-3;
	for(iy = 0 ; iy < pm->ny ; iy++) {
		printf("\r *** Making data [%03d/%03d] ***", iy+1, pm->ny);
		y = (pm->ny/2-iy)*20./pm->ny;
		for(ix = 0 ; ix < pm->nx ; ix++) {
			if(pm->iro[iy*pm->nx+ix] == 0.0)  continue;
			x = (ix-pm->nx/2)*20./pm->nx;
			a = 1.-exp(-pm->tr/pm->it1[iy*pm->nx+ix]);
			for(i = 0 ; i < pm->ny ; i++) {
				m = pm->ny/2-i;
				for(j = 0 ; j < pm->nx ; j++) {
					k = j-pm->nx/2;
					b = exp(-1.*fabs(k*1e-4)/0.003)*exp(-1.*(k*1e-4+pm->te*1e-3)/(pm->it2[iy*pm->nx+ix]*1e-3)); 
					th = g0*g1*(x*k+y*m);
					pm->imr[i*pm->nx+j] += (float)(pm->iro[iy*pm->nx+ix]*a*cos(th)*b);
					pm->imi[i*pm->nx+j] += (float)(pm->iro[iy*pm->nx+ix]*a*sin(th)*b);
				}
			}
		}
	}
	printf("\n");
}

void complex_conjucate_1D(float *imr, float *imi, int nx, int ny)
{
	int    i, j;
	float  sr, si;
	void   fswap(float *, float *);

	for(i = 0 ; i < ny ; i++) {
		sr = imr[i*nx+nx/2];
		si = imi[i*nx+nx/2];
		for(j = 0 ; j < nx/2 ; j++) {
			imr[i*nx+j+nx/2] = imr[i*nx+j];
			imi[i*nx+j+nx/2] = imi[i*nx+j];
		}
		imr[i*nx] = sr;
		imi[i*nx] = -si;
	}
	for(j = 1 ; j < nx/2 ; j++) {
		imr[j] =  imr[nx-j];
		imi[j] = -imi[nx-j];
	}
	for(i = 1 ; i < nx/2 ; i++) {
		fswap(imr+i*nx, imr+(nx-i)*nx);
		fswap(imi+i*nx, imi+(nx-i)*nx);
	}
	for(i = 1 ; i < ny ; i++) {
		for(j = 1 ; j < nx/2 ; j++) {
			imr[i*nx+j] =  imr[(nx-i)*nx+nx-j];
			imi[i*nx+j] = -imi[(nx-i)*nx+nx-j];
		}
	}
}
