/*
	File : recon_correct_propeller.c
*/

/* --- vO̐ ---

12́@PROPELLER MRI
4߁@t@gpV~[Vƕ␳

@ʑ̂̕sړƉ]ړ܂񂾓_PROPELLER MRI
vf[^͂Cړ̕␳čč\vOD

́F
  1Dvf[^ij̃t@CiPx_^j
  2Dvf[^ij̃t@CiPx_^j
  3Dvf[^x̃}gNXTCYi2̃xLj
  4Dvf[^ỹ}gNXTCYi2̃xLj
  5DuCh̐
  6Dč\ꂽ摜̏o̓t@CiPx_^j
  7Dč\摜x̃}gNXTCYi2̃xLj
  8Dč\摜ỹ}gNXTCYi2̃xLj

óF
  ␳čč\ꂽ_摜@in0.imgj

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  9     /* number of parameters + 1 */

typedef struct {
	char   f1[50]; /* input source file  name */
	char   f2[50]; /* input source file  name */
	int    nx;     /* number of matrix (x) */
	int    ny;     /* number of matrix (y) */
	int    nb;     /* number of blade  (b) */
	char   f3[50]; /* output reconstruction file  name */
	int    mx;     /* number of matrix (x) */
	int    my;     /* number of matrix (y) */
	float  *sor;   /* Source data (real)  */
	float  *soi;   /* Source data (imaginary) */
	float  *imr;   /* Image data (real)  */
	float  *imi;   /* Image data (imaginary) */
} Param;

char *menu[PN] = {
	"MRI rotation correct for point source (PROPELLER MRI)",
	"Input source file name (Re)    <.img> ",
	"Input source file name (Im)    <.img> ",
	"Number of matrix  (nx)       ",
	"Number of matrix  (ny)       ",
	"Number of blade   (nb)       ",
	"Output reconstruction file name <.img> ",
	"Number of matrix  (mx)       ",
	"Number of matrix  (my)       ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void make_point_propeller(Param *);
void to_matrix1(int, float *, float *, Param *);
void to_matrix(float *, Param *);
void move(float *, int, int, double, double);
void rotate(float *, float *, int, int, double);
double calc_ang(float *, float *, int, int, int, double, int *, int *);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void FFT_2D(int, float *, float *, int, int);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Pring Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "n0r.img");
	sprintf( pm->f2, "n0i.img");
	pm->nx = 128;
	pm->ny = 25;
	pm->nb = 8;
	sprintf( pm->f3, "n0.img");
	pm->mx = 128;
	pm->my = 128;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f2 );
		if(*gets(dat) != '\0')  strcpy(pm->f2, dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->nb );
		if(*gets(dat) != '\0')  pm->nb = atoi(dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f3 );
		if(*gets(dat) != '\0')  strcpy(pm->f3, dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->mx );
		if(*gets(dat) != '\0')  pm->mx = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->my );
		if(*gets(dat) != '\0')  pm->my = atoi(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) strcpy( pm->f2, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->nb = atoi( argv[i++] );
		if((argc--) > 1) strcpy( pm->f3, argv[i++] );
		if((argc--) > 1) pm->mx = atoi( argv[i++] );
		if((argc--) > 1) pm->my = atoi( argv[i++] );
	}
	else {
		usage(argc, argv);
	}
}

void main(int argc, char *argv[] )
{
	char    fi[50];
	int     i, xmax, ymax;
	float   *ang, *rot;
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->sor = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nb*sizeof(float));
	pm->soi = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nb*sizeof(float));
	pm->imr = (float *)malloc((unsigned long)pm->mx*pm->my*pm->nb*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->mx*pm->my*pm->nb*sizeof(float));
	rot = (float *)malloc((unsigned long)pm->mx*pm->my*sizeof(float));
	ang = (float *)malloc((unsigned long)pm->nb*sizeof(float));

	printf(" *** Read Source Image ***\n");
	read_data(pm->f1, pm->sor, pm->nx*pm->ny*pm->nb);
	read_data(pm->f2, pm->soi, pm->nx*pm->ny*pm->nb);

	printf(" *** To Matrix ***\n");
	to_matrix1(0, pm->imr, pm->imi, pm);
	write_data("n00r.img", pm->imr, pm->mx*pm->my);
	write_data("n00i.img", pm->imi, pm->mx*pm->my);
	FFT_2D(-1, pm->imr, pm->imi, pm->mx, pm->my);
	write_data("n00rr.img", pm->imr, pm->mx*pm->my);

	ang[0] = 0;
	for(i = 1 ; i < pm->nb ; i++) {
		to_matrix1(i, pm->imr+i*pm->mx*pm->my, pm->imi+i*pm->mx*pm->my, pm);
		sprintf(fi, "n0%dr.img", i);
		write_data(fi, pm->imr+i*pm->mx*pm->my, pm->mx*pm->my);
		sprintf(fi, "n0%di.img", i);
		write_data(fi, pm->imi+i*pm->mx*pm->my, pm->mx*pm->my);

		FFT_2D(-1, pm->imr+i*pm->mx*pm->my, pm->imi+i*pm->mx*pm->my, pm->mx, pm->my);
		sprintf(fi, "n0%drr.img", i);
		write_data(fi, pm->imr+i*pm->mx*pm->my, pm->mx*pm->my);

		ang[i] = (float)calc_ang(pm->imr+i*pm->mx*pm->my, pm->imr, pm->mx, pm->my, 2, 0.01, &xmax, &ymax); 

		xmax -= pm->mx/2;
		ymax = pm->my/2-ymax;
		fprintf(stderr, "[%d] xmax = %d, ymax = %d \n", i, xmax, ymax);

		move(pm->imr+i*pm->mx*pm->my, pm->mx, pm->my, (float)xmax, (float)ymax);
		move(pm->imi+i*pm->mx*pm->my, pm->mx, pm->my, (float)xmax, (float)ymax);
		rotate(pm->imr+i*pm->mx*pm->my, pm->imr+i*pm->mx*pm->my, pm->mx, pm->my, ang[i]);
		rotate(pm->imi+i*pm->mx*pm->my, pm->imi+i*pm->mx*pm->my, pm->mx, pm->my, ang[i]);
		FFT_2D(1, pm->imr+i*pm->mx*pm->my, pm->imi+i*pm->mx*pm->my, pm->mx, pm->my);
	}

	to_matrix(ang, pm);
	write_data("n300r.img", pm->imr, pm->mx*pm->my);
	write_data("n300i.img", pm->imi, pm->mx*pm->my);

	FFT_2D(-1, pm->imr, pm->imi, pm->mx, pm->my);

	printf(" *** Write Image data   ***\n");
	write_data(pm->f3, pm->imr, pm->mx*pm->my);

	free(pm->sor);
	free(pm->soi);
	free(pm->imr);
	free(pm->imi);
	free(rot);
	free(ang);
	free(pm);
}

void to_matrix1(int n, float *imr, float *imi, Param *pm)
{
	int     i, j, ix, iy, xj, yi;
	double  x, y, dx, dy, an;

	for(i = 0 ; i < pm->mx*pm->my ; i++)
		imr[i] = imi[i] = 0;
	an = PI*n/pm->nb;
	for(i = 0 ; i < pm->my ; i++) {
		yi = pm->my/2-i;
		for(j = 0 ; j < pm->mx ; j++) {
			xj = j - pm->mx/2;
			x =  xj*cos(an)+yi*sin(an);
			y = -xj*sin(an)+yi*cos(an);
			x = x + pm->nx/2;
			y = pm->ny/2-y;
			ix = (int)x;
			iy = (int)y;
			if(ix < 0 || ix+1 >= pm->nx || iy < 0 || iy+1 >= pm->ny)
				continue;
			dx = x - ix;
			dy = y - iy;
			imr[i*pm->mx+j] = 
				(float)((1-dy)*(1-dx)*pm->sor[n*pm->ny*pm->nx+ iy   *pm->nx+ix  ]
				      + (1-dy)*   dx *pm->sor[n*pm->ny*pm->nx+ iy   *pm->nx+ix+1]
				      +    dy *(1-dx)*pm->sor[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix  ]
				      +    dy *   dx *pm->sor[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix+1]);
			imi[i*pm->mx+j] = 
				(float)((1-dy)*(1-dx)*pm->soi[n*pm->ny*pm->nx+ iy   *pm->nx+ix  ]
				      + (1-dy)*   dx *pm->soi[n*pm->ny*pm->nx+ iy   *pm->nx+ix+1]
				      +    dy *(1-dx)*pm->soi[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix  ]
				      +    dy *   dx *pm->soi[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix+1]);
		}
	}
}

void to_matrix(float *ang, Param *pm)
{
	int     i, j, n, ix, iy, xj, yi;
	int     *pn;
	double  x, y, an;

	pn = (int *)malloc((unsigned long)pm->mx*pm->my*sizeof(int));

	for(i = 0 ; i < pm->mx*pm->my ; i++)
		pm->imr[i] = pm->imi[i] = (float)pn[i] = 0;
	for(n = 0 ; n < pm->nb ; n++) {
		an = PI*n/pm->nb-ang[n]*PI/180.;
		for(i = 0 ; i < pm->my ; i++) {
			yi = pm->my/2-i;
			for(j = 0 ; j < pm->mx ; j++) {
				xj = j - pm->mx/2;
				x =  xj*cos(an)+yi*sin(an);
				y = -xj*sin(an)+yi*cos(an);
				x = x + pm->nx/2;
				y = pm->ny/2-y;
				ix = (int)x;
				iy = (int)y;
				if(ix < 0 || ix+1 >= pm->nx || iy < 0 || iy+1 >= pm->ny)
					continue;
				pm->imr[i*pm->mx+j] += pm->imr[n*pm->my*pm->mx+i*pm->mx+j];
				pm->imi[i*pm->mx+j] += pm->imi[n*pm->my*pm->mx+i*pm->mx+j];
				pn[i*pm->mx+j] ++;
			}
		}
	}
	for(i = 0 ; i < pm->mx*pm->my ; i++)
		if(pn[i] != 0) {
			pm->imr[i] /= pn[i];
			pm->imi[i] /= pn[i];
		}
	free(pn);
}

void move(float *img, int nx, int ny, double dx, double dy)
{
	int     i, j, i0, i1, j0, j1;
	double  x0, y0;
	float   *ima;

	ima = (float *)malloc((unsigned long)nx*ny*sizeof(float));
	for(i = 0 ; i < nx*ny ; i++)
		ima[i] = 0;

	for(i = 0 ; i < ny ; i++) {
		y0 = i+dy; // dy͋̕tɂiÿړj
		i0 = (int)y0;
		i1 = i0+1;
		if(i0 < 0 || i1 > ny-1) continue;
		for(j = 0 ; j < nx ; j++) {
			x0 = j-dx; // iẍړj
			j0 = (int)x0;
			j1 = j0+1;
			if(j0 < 0 || j1 > nx-1) continue;
			ima[i*nx+j] = (float)((j1-x0)*(i1-y0)*img[i0*nx+j0]
			                    + (x0-j0)*(i1-y0)*img[i0*nx+j1]
			                    + (j1-x0)*(y0-i0)*img[i1*nx+j0]
			                    + (x0-j0)*(y0-i0)*img[i1*nx+j1]);
		}
	}
	for(i = 0 ; i < nx*ny ; i++)
		img[i] = ima[i];
	free(ima);
}

void correlation_f(float *rre, float *rim, float *fre, float *fim, float *gre, float *gim, int n) 
{
	int   i;

	for(i = 0 ; i < n ; i++) {
		rre[i] = fre[i]*gre[i]+fim[i]*gim[i];
		rim[i] = fim[i]*gre[i]-fre[i]*gim[i];
	}
}

void rotate(float *rot, float *img, int nx, int ny, double th)
{
	int     i, j, i0, i1, j0, j1;
	double  x0, y0, si, co;
	float   *ro;

	ro = (float *)malloc((unsigned long)nx*ny*sizeof(float));

	for(i = 0 ; i < nx*ny ; i++)
		ro[i] = 0;

	si = sin(th*PI/180.);
	co = cos(th*PI/180.);

	for(i = 0 ; i < ny ; i++) {
		for(j = 0 ; j < nx ; j++) {
			x0 = (j-nx/2)*co-(ny/2-i)*si+nx/2; // ]ixj
			j0 = (int)x0;
			j1 = j0+1;
			if(j0 < 0 || j1 > nx-1) continue;
			y0 = ny/2-(j-nx/2)*si-(ny/2-i)*co; // ]iyj
			i0 = (int)y0;
			i1 = i0+1;
			if(i0 < 0 || i1 > ny-1) continue;
			ro[i*nx+j] = (float)((j1-x0)*(i1-y0)*img[i0*nx+j0]
			                   + (x0-j0)*(i1-y0)*img[i0*nx+j1]
			                   + (j1-x0)*(y0-i0)*img[i1*nx+j0]
			                   + (x0-j0)*(y0-i0)*img[i1*nx+j1]);
		}
	}
	for(i = 0 ; i < nx*ny ; i++)
		rot[i] = ro[i];
	free(ro);
}

float max_value(float *img, int nx, int ny, int *x, int *y)
{
	int   i, j;
	float max;

	max = img[0];
	*x = *y = 0;
	for(i = 0 ; i < ny ; i++) {
		for(j = 0 ; j < nx ; j++) {
			if(img[i*nx+j] > max) {
				max = img[i*nx+j];
				*x = j;
				*y = i;
			}
		}
	}
	return max;
}

double max_ang(float *f0r, float *p1r, float *p1i, int nx, int ny, double t, double ang, double th, float *max, int *x, int *y) 
{
	int    i, xx, yy;
	float  m;
	float  *f1r, *f1i, *f2r, *f2i;

	f1r = (float *)malloc((unsigned long)nx*ny*sizeof(float));
	f1i = (float *)malloc((unsigned long)nx*ny*sizeof(float));
	f2r = (float *)malloc((unsigned long)nx*ny*sizeof(float));
	f2i = (float *)malloc((unsigned long)nx*ny*sizeof(float));

	rotate(f1r, f0r, nx, ny, t);
	for(i = 0 ; i < nx*ny ; i++)
		f1i[i] = 0;
	FFT_2D(1, f1r, f1i, nx, ny);
	correlation_f(f2r, f2i, p1r, p1i, f1r, f1i, nx*ny);
	FFT_2D(-1, f2r, f2i, nx, ny);
	m = max_value(f2r, nx, ny, &xx, &yy);
	//m = f2r[ny/2*nx+nx/2];
	if(m > *max) {
		*max = m;
		*x = xx;
		*y = yy;
		ang = t;
	}
	free(f1r);
	free(f1i);
	free(f2r);
	free(f2i);
	return  ang;
}

double calc_ang(float *f0r, float *f1r, int nx, int ny, int dp, double th, int *x, int *y) 
{
	int     i, k;
	float   max;
	float   *f0i, *f1i, *f2r, *f2i;
	double  t, ang, st, dt;

	f0i = (float *)malloc((unsigned long)nx*ny*sizeof(float));
	f1i = (float *)malloc((unsigned long)nx*ny*sizeof(float));
	f2r = (float *)malloc((unsigned long)nx*ny*sizeof(float));
	f2i = (float *)malloc((unsigned long)nx*ny*sizeof(float));

	for(i = 0 ; i < nx*ny ; i++)
		f1i[i] = f0i[i] = 0;

	write_data("cor0.img",f1r,nx*ny);
	FFT_2D(1, f0r, f0i, nx, ny);
	FFT_2D(1, f1r, f1i, nx, ny);

	correlation_f(f2r, f2i, f1r, f1i, f0r, f0i, nx*ny);
	FFT_2D(-1, f2r, f2i, nx, ny);
	max = max_value(f2r, nx, ny, x, y);
	//max = f2r[ny/2*nx+nx/2];
	ang = 0;
	free(f2r);
	free(f2i);
	printf(" Ang = %f, max=%f\n", ang, max);

	FFT_2D(-1, f0r, f0i, nx, ny);

	dt = 1;
	for(t = 1 ; t < 360. ; t+=dt) {
		ang = max_ang(f0r, f1r, f1i, nx, ny, t, ang, th, &max, x, y);
	}
	printf(" Ang = %f, max=%f\n", ang, max);

	for(k = -1 ; k < dp ; k++) {
		st = ang;
		dt *= 0.1;
		for(t = st-dt*10 ; t < st+dt*10 ; t+=dt) {
			ang = max_ang(f0r, f1r, f1i, nx, ny, t, ang, th, &max, x, y);
		}
	printf(" Ang = %f, max=%f\n", ang, max);
	}

	if(ang > 180.) ang -= 360;
	printf("\nAngle    [ %f degree ] \n", ang);

	FFT_2D(-1, f1r, f1i, nx, ny);

	free(f0i);
	free(f1i);

	return ang;
}
