/*
	File : recon_motion_propeller.c
*/

/* --- vO̐ ---

12́@PROPELLER MRI
4߁@t@gpV~[Vƕ␳

@lt@gƁuv͂āCPROPELLER MRǏvf[^쐬C
2Fourierϊ̈̃f[^ɕϊčč\vOD

́F
  1Dvf[^ijp̏o̓t@CiPx_^j
  2Dvf[^ijp̏o̓t@CiPx_^j
  3Dvf[^x̃}gNXTCYi2̃xLj
  4Dvf[^ỹ}gNXTCYi2̃xLj
  5DuCh̐
  6Dč\ꂽ摜̏o̓t@CiPx_^j
  7Dč\摜x̃}gNXTCYi2̃xLj
  8Dč\摜ỹ}gNXTCYi2̃xLj
  9. lt@g̃t@CiPx_^j
 10. x̑icm/sj
 11. y̑icm/sj
 12. ̊px̑idegree/sj

óF
  vf[^̎̉摜in0r.imgj
  vf[^̋̉摜in0i.imgj
  2t[Gϊ̈ɕϊꂽ̉摜in00r.imgj
  2t[Gϊ̈ɕϊꂽ̉摜in00i.imgj
  č\ꂽ_摜@in0.imgj

Kvȃt@CF
  io.c        i摜̓o͊֐̃t@Cj
  fft.c       i1,2,3FFT֐̃t@Cj

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  PI  3.14159265358979
#define  PN  13     /* number of parameters + 1 */

typedef struct {
	char    f1[50]; /* output source file  name */
	char    f2[50]; /* output source file  name */
	int     nx;     /* number of matrix (x) */
	int     ny;     /* number of matrix (y) */
	int     nb;     /* number of blade  (b) */
	char    f3[50]; /* output reconstruction file  name */
	int     mx;     /* number of matrix (x) */
	int     my;     /* number of matrix (y) */
	char    f4[50]; /* input source file  name */
	double  px;     /* x-coordinate of point source */
	double  py;     /* y-coordinate of point source */
	double  vl;     /* value of point source */
	double  vx;     /* motion  velocity (x) */
	double  vy;     /* motion  velocity (y) */
	double  va;     /* angular velocity (a) */
	float   *sor;   /* Source data (real)  */
	float   *soi;   /* Source data (imaginary) */
	float   *imr;   /* Image data (real)  */
	float   *imi;   /* Image data (imaginary) */
} Param;

char *menu[PN] = {
	"MRI reconstruction for numerical phantom (PROPELLER MRI)",
	"Output source file name (Re)    <.img> ",
	"Output source file name (Im)    <.img> ",
	"Number of matrix  (nx)       ",
	"Number of matrix  (ny)       ",
	"Number of blade   (nb)       ",
	"Output reconstruction file name <.img> ",
	"Number of matrix  (mx)       ",
	"Number of matrix  (my)       ",
	"Input source file name          <.img> ",
	"motion velocity  (vx)        ",
	"motion velocity  (vy)        ",
	"angular velocity (va)        ",
	};

void read_data(char *, float *, int);
void write_data(char *, float *, int);
void make_recon_propeller(Param *);
void make_point_propeller(Param *);
void to_matrix(Param *);
void FFTInit(int , float *, float *, unsigned short *);
void FFT(int, int, float *, float *, float *, float *, unsigned short *);
void FFT_2D(int, float *, float *, int, int);

void usage(int argc, char **argv)
{
	int   i;

	fprintf( stderr,"\nUSAGE:\n");
	fprintf( stderr,"\nNAME\n");
	fprintf( stderr,"\n  %s - %s\n", argv[0], menu[0]);
	fprintf( stderr,"\nSYNOPSIS\n");
	fprintf( stderr,"\n  %s [-h] parameters...\n", argv[0]);
	fprintf( stderr,"\nPARAMETERS\n");
	for(i = 1 ; i < PN ; i++)
	  fprintf( stderr,"\n %3d. %s\n", i, menu[i]);
	fprintf( stderr,"\n");
	fprintf( stderr,"\nFLAGS\n");
	fprintf( stderr,"\n  -h  Pring Usage (this comment).\n");
	fprintf( stderr,"\n");
	exit(1);
}

void getparameter(int argc, char **argv, Param *pm)
{
	int   i;
	char  dat[256];

	/* default parameter value */
	sprintf( pm->f1, "n0r.img");
	sprintf( pm->f2, "n0i.img");
	pm->nx = 128;
	pm->ny = 25;
	pm->nb = 8;
	sprintf( pm->f3, "n0.img");
	pm->mx = 128;
	pm->my = 128;
	sprintf( pm->f4, "shepp3n.img");
	pm->vx = 0.3;
	pm->vy = 0.2;
	pm->va = 3;

	i = 0;
	if( argc == 1+i ) {
		fprintf( stdout, "\n%s\n\n", menu[i++] );
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f1 );
		if(*gets(dat) != '\0')  strcpy(pm->f1, dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f2 );
		if(*gets(dat) != '\0')  strcpy(pm->f2, dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->nx );
		if(*gets(dat) != '\0')  pm->nx = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->ny );
		if(*gets(dat) != '\0')  pm->ny = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->nb );
		if(*gets(dat) != '\0')  pm->nb = atoi(dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f3 );
		if(*gets(dat) != '\0')  strcpy(pm->f3, dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->mx );
		if(*gets(dat) != '\0')  pm->mx = atoi(dat);
		fprintf( stdout, " %s [%d] :", menu[i++], pm->my );
		if(*gets(dat) != '\0')  pm->my = atoi(dat);
		fprintf( stdout, " %s [%s] :", menu[i++], pm->f4 );
		if(*gets(dat) != '\0')  strcpy(pm->f4, dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->vx );
		if(*gets(dat) != '\0')  pm->vx = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->vy );
		if(*gets(dat) != '\0')  pm->vy = atof(dat);
		fprintf( stdout, " %s [%f] :", menu[i++], pm->va );
		if(*gets(dat) != '\0')  pm->va = atof(dat);
	}
	else if ( argc == PN+i ) {
		fprintf( stderr, "\n%s [%s]\n", argv[i++], menu[0] );
		if((argc--) > 1) strcpy( pm->f1, argv[i++] );
		if((argc--) > 1) strcpy( pm->f2, argv[i++] );
		if((argc--) > 1) pm->nx = atoi( argv[i++] );
		if((argc--) > 1) pm->ny = atoi( argv[i++] );
		if((argc--) > 1) pm->nb = atoi( argv[i++] );
		if((argc--) > 1) strcpy( pm->f3, argv[i++] );
		if((argc--) > 1) pm->mx = atoi( argv[i++] );
		if((argc--) > 1) pm->my = atoi( argv[i++] );
		if((argc--) > 1) strcpy( pm->f4, argv[i++] );
		if((argc--) > 1) pm->vx = atof( argv[i++] );
		if((argc--) > 1) pm->vy = atof( argv[i++] );
		if((argc--) > 1) pm->va = atof( argv[i++] );
	}
	else {
		usage(argc, argv);
	}
}

void main(int argc, char *argv[] )
{
	Param   *pm;

	pm = (Param *)malloc(sizeof(Param));
	getparameter(argc, argv, pm);

	pm->sor = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nb*sizeof(float));
	pm->soi = (float *)malloc((unsigned long)pm->nx*pm->ny*pm->nb*sizeof(float));
	pm->imr = (float *)malloc((unsigned long)pm->mx*pm->my*sizeof(float));
	pm->imi = (float *)malloc((unsigned long)pm->mx*pm->my*sizeof(float));

	printf(" *** Making Point Image ***\n");
	read_data(pm->f4, pm->imr, pm->mx*pm->my);
	make_recon_propeller(pm);
	write_data(pm->f1, pm->sor, pm->nx*pm->ny*pm->nb);
	write_data(pm->f2, pm->soi, pm->nx*pm->ny*pm->nb);

	printf(" *** To Matrix ***\n");
	to_matrix(pm);

	write_data("n00r.img", pm->imr, pm->mx*pm->my);
	write_data("n00i.img", pm->imi, pm->mx*pm->my);

	printf(" *** 2D-IFT  ***\n");
	FFT_2D(-1, pm->imr, pm->imi, pm->mx, pm->my);

	printf(" *** Write Image data   ***\n");
	write_data(pm->f3, pm->imr, pm->mx*pm->my);

	free(pm->sor);
	free(pm->soi);
	free(pm->imr);
	free(pm->imi);
	free(pm);
}

void make_recon_propeller(Param *pm)
{
	int      i, j;

	for(i = 0 ; i < pm->nx*pm->ny*pm->nb ; i++)
		pm->sor[i] = pm->soi[i] = 0;
	for(i = 0 ; i < pm->my ; i++) {
		fprintf(stderr, "\r --- motion [%03d/%03d] ---", i+1, pm->my);
		pm->py = (pm->my/2-i)*20./pm->my;
		for(j = 0 ; j < pm->mx ; j++) {
			if((pm->vl = pm->imr[i*pm->mx+j]) == 0.0) continue;
			pm->px = (j-pm->mx/2)*20./pm->mx;
			make_point_propeller(pm);
		}
	}
	fprintf(stderr,"\n");
}

void make_point_propeller(Param *pm)
{
	int	     i, j, n;
	double   an, th, g, x0, y0, xn, yn;

	g = 2*PI*42.58*1e6*0.12*1e-4*0.1*1e-3;
	for(n = 0 ; n < pm->nb ; n++) {
		an = -n*pm->va*PI/180.;
		xn =  pm->px*cos(an)+pm->py*sin(an);
		yn = -pm->px*sin(an)+pm->py*cos(an);
		xn += n*pm->vx;
		yn += n*pm->vy;
		an = PI*n/pm->nb;
		x0 =  xn*cos(an)+yn*sin(an);
		y0 = -xn*sin(an)+yn*cos(an);
		for(i = 0 ; i < pm->ny ; i++) { // before movement
			for(j = 0 ; j < pm->nx ; j++) {
				th = g*(x0*(j-pm->nx/2)+y0*(pm->ny/2-i));
				pm->sor[n*pm->nx*pm->ny+i*pm->nx+j] += (float)(pm->vl*cos(th));
				pm->soi[n*pm->nx*pm->ny+i*pm->nx+j] += (float)(pm->vl*sin(th));
			}
		}
	}
}

void to_matrix(Param *pm)
{
	int     i, j, n, ix, iy, xj, yi;
	int     *pn;
	double  x, y, dx, dy, an;

	pn = (int *)malloc((unsigned long)pm->mx*pm->my*sizeof(int));

	for(i = 0 ; i < pm->mx*pm->my ; i++)
		pm->imr[i] = pm->imi[i] = (float)pn[i] = 0;
	for(n = 0 ; n < pm->nb ; n++) {
		an = PI*n/pm->nb;
		for(i = 0 ; i < pm->my ; i++) {
			yi = pm->my/2-i;
			for(j = 0 ; j < pm->mx ; j++) {
				xj = j - pm->mx/2;
				x =  xj*cos(an)+yi*sin(an);
				y = -xj*sin(an)+yi*cos(an);
				x = x + pm->nx/2;
				y = pm->ny/2-y;
				ix = (int)x;
				iy = (int)y;
				if(ix < 0 || ix+1 >= pm->nx || iy < 0 || iy+1 >= pm->ny)
					continue;
				dx = x - ix;
				dy = y - iy;
				pm->imr[i*pm->mx+j] += 
					(float)((1-dy)*(1-dx)*pm->sor[n*pm->ny*pm->nx+ iy   *pm->nx+ix  ]
					      + (1-dy)*   dx *pm->sor[n*pm->ny*pm->nx+ iy   *pm->nx+ix+1]
					      +    dy *(1-dx)*pm->sor[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix  ]
					      +    dy *   dx *pm->sor[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix+1]);
				pm->imi[i*pm->mx+j] += 
					(float)((1-dy)*(1-dx)*pm->soi[n*pm->ny*pm->nx+ iy   *pm->nx+ix  ]
					      + (1-dy)*   dx *pm->soi[n*pm->ny*pm->nx+ iy   *pm->nx+ix+1]
					      +    dy *(1-dx)*pm->soi[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix  ]
					      +    dy *   dx *pm->soi[n*pm->ny*pm->nx+(iy+1)*pm->nx+ix+1]);
				pn[i*pm->mx+j] ++;
			}
		}
	}
	for(i = 0 ; i < pm->mx*pm->my ; i++)
		if(pn[i] != 0) {
			pm->imr[i] /= pn[i];
			pm->imi[i] /= pn[i];
		}
	free(pn);
}
